Components.utils.importModule("chrome://phone/content/utils/class.js");

function fixupPhoneURI(uriSpec) {
  var fixup = Components.classes["@mozilla.org/phone/phone-urifixup;1"].
    getService(Components.interfaces.nsIURIFixup);
  try { return fixup.createFixupURI(uriSpec, 0); }
  catch(e) { return null; }
}
function fixupWebURI(uriSpec) {
  var fixup = Components.classes["@mozilla.org/docshell/urifixup;1"].
    getService(Components.interfaces.nsIURIFixup);
  try { return fixup.createFixupURI(uriSpec, 0); }
  catch(e) { return null; }
}
function fixupMailURI(uriSpec) {
  var fixup = Components.classes["@mozilla.org/phone/mail-urifixup;1"].
    getService(Components.interfaces.nsIURIFixup);
  try { return fixup.createFixupURI(uriSpec, 0); }
  catch(e) { return null; }
}

function Service() {
  var session = Components.classes["@mozilla.org/addressbook/services/session;1"].
    getService(Components.interfaces.nsIAddrBookSession);
  session.addAddressBookListener(this, -1);

  var manager = Components.classes["@mozilla.org/phone/sip-ua-manager;1"].
    getService(Components.interfaces.phISipUserAgentManager);
  // some day null should be replaced by a reference to phISipServer/nsIMsgIncomingServer
  this.sipUserAgent = manager.getClient(null);

  this.crawlAll();
}
var service = Service.prototype = new Component("phIAddressLookupService", "nsIAbListener");

service.sipUserAgent = null;

// nsIAbListener handlers:

service.onItemAdded = function(parentDir, item) {
  // crawl only new card
  var directory = parentDir.QueryInterface(Components.interfaces.nsIAbDirectory),
    card = item.QueryInterface(Components.interfaces.nsIAbCard);
  this.crawlDirectory(directory, card);
};
service.onItemPropertyChanged = function(item, property, oldValue, newValue) {
  // recrawl complete address book when a relevant property is changed
  // (as former overlapped properties could have become visible now)
  if(
    property == null ||
    property == "PrimarySIP" ||
    property == "SecondarySIP" ||
    property == "PrimaryEmail" ||
    property == "SecondaryEmail" ||
    property == "WorkPhone" ||
    property == "HomePhone" ||
    property == "WebPage1" ||
    property == "WebPage2"
  )
    this.crawlAll();
};
service.onItemRemoved = function() {
  // recrawl whole address book (as former overlapped properties could have become visible now)
  this.crawlAll();
};


service.crawlAll = function() {
  this.items = {};
  this.identities = [];

  var rootURI = "moz-abdirectory://";
  var rdf = Components.classes["@mozilla.org/rdf/rdf-service;1"].
    getService(Components.interfaces.nsIRDFService);
  var rootDirectory = rdf.GetResource(rootURI).
    QueryInterface(Components.interfaces.nsIAbDirectory);

  this.crawlDirectory(rootDirectory);
};
service.crawlDirectory = function(directory) {
  var childDirectories = directory.childNodes;
  while(childDirectories.hasMoreElements()) {
    var childDirectory = childDirectories.getNext();
    if(!childDirectory.isRemote)
      this.crawlDirectory(childDirectory.QueryInterface(Components.interfaces.nsIAbDirectory));
  }

  try { var cards = directory.childCards; }
  catch(e) { var cards = null; }
  if(cards != null) {
    var hasMoreElements = true;
    try { cards.first(); }
    catch(e) { hasMoreElements = false; }
    while(hasMoreElements) {
      var card = cards.currentItem().QueryInterface(Components.interfaces.nsIAbCard);
      this.crawlCard(directory, card);
      try { cards.next(); }
      catch(e) { hasMoreElements = false; }
    }
  }
};
service.crawlCard = function(directory, card) {
  var primaryMailURI = card.primaryEmail? fixupMailURI(card.primaryEmail): null;
  var secondaryMailURI = card.secondaryEmail? fixupMailURI(card.secondaryEmail): null;

  var primarySIP = directory.getValueForCard(card, "PrimarySIP");
  var secondarySIP = directory.getValueForCard(card, "SecondarySIP");
  var primaryPhoneURI = primarySIP? fixupPhoneURI(primarySIP): null;
  var secondaryPhoneURI = secondarySIP? fixupPhoneURI(secondarySIP): null;

  var cellularPhoneURI = card.cellularNumber? fixupPhoneURI(card.cellularNumber): null;
  var homePhoneURI = card.homePhone? fixupPhoneURI(card.homePhone): null;
  var workPhoneURI = card.workPhone? fixupPhoneURI(card.workPhone): null;

  var workWebURI = card.webPage1? fixupWebURI(card.webPage1): null;
  var homeWebURI = card.webPage2? fixupWebURI(card.webPage2): null;

  var id = this.createIdentity();
  id.displayName = card.displayName;
  if(primaryMailURI) this.registerURI(primaryMailURI, id);
  if(secondaryMailURI) this.registerURI(secondaryMailURI, id);
  if(primaryPhoneURI) this.registerURI(primaryPhoneURI, id);
  if(secondaryPhoneURI) this.registerURI(secondaryPhoneURI, id);
  if(cellularPhoneURI) this.registerURI(cellularPhoneURI, id);
  if(homePhoneURI) this.registerURI(homePhoneURI, id);
  if(workPhoneURI) this.registerURI(workPhoneURI, id);
  if(workWebURI) this.registerURI(workWebURI, id);
  if(homeWebURI) this.registerURI(homeWebURI, id);
};

service.items = null;
service.registerURI = function(uri, identity) {
  var uriSpec = uri.spec, items = this.items;
  if(!items.hasOwnProperty(uriSpec)) items[uriSpec] = [];
  var item = {identity: identity, uri: uri};
  items[uriSpec].push(item);
  identity.uris[uriSpec] = uri;

  // see RFC 3856 section 4
  if(uri.schemeIs("sip") || uri.schemeIs("sips") || uri.schemeIs("pres")) {
    this.sipUserAgent.subscribe(uri);
  }
};

function Identity() { this.uris = {}; };
service.identities = null;
service.createIdentity = function() {
  var index = this.identities.length;
  var identity = this.identities[index] = new Identity;
  identity.index = index;
  return identity;
};

service.getAlternativeURIs = function(originalURISpec) {
  var uris = Components.classes["@mozilla.org/array;1"].
    createInstance(Components.interfaces.nsIMutableArray);
  if(!this.items.hasOwnProperty(originalURISpec)) return uris;

  // first write any found URIs to a map to avoid redundancy
  var uriMap = {};
  var uriItems = this.items[originalURISpec];
  for(var i = 0; i < uriItems.length; i++) {
    var identity = uriItems[i].identity;
    for(var uriSpec in identity.uris) {
      var uri = identity.uris[uriSpec];
      uriMap[uriSpec] = uri;
    }
  }
  // then copy unique URIs to the nsIArray
  for(var uriSpec in uriMap) {
    uris.appendElement(uriMap[uriSpec], false);
  }
  return uris;
};

service.getAlternativeURISpec = function(originalURISpec, scheme) {
  if(!this.items.hasOwnProperty(originalURISpec)) return null;
  var uriItems = this.items[originalURISpec];
  for(var i = 0; i < uriItems.length; i++) {
    var identity = uriItems[i].identity;
    for(var uriSpec in identity.uris) {
      var uri = identity.uris[uriSpec];
      if(uri.schemeIs(scheme)) return uri.spec;
    }
  }
  return null;
};

service.isOnline = function(uriSpec) {
  var uris = this.getAlternativeURIs(uriSpec);
  for(var online = false, i = 0; i < uris.length; i++) {
    var uri = uris.queryElementAt(i, Components.interfaces.nsIURI);
    if(uri.schemeIs("sip") || uri.schemeIs("sips"))
      online = true;
  }
  return online;
};

function Result() {
  this.addresses = Components.classes["@mozilla.org/array;1"].
    createInstance(Components.interfaces.nsIMutableArray);
}
Result.prototype.push = function(item) {
  var address = Components.classes["@mozilla.org/phone/address;1"].
    createInstance(Components.interfaces.phIAddress);
  address.uri = item.uri;
  address.displayName = item.identity.displayName;
  this.addresses.appendElement(address, false);
};
service.matchAddresses = function(query) {
  query = query.replace(/^\s+/, "");
  query = query.replace(/\s+$/, "");
  query = query.toLowerCase();
  if(!query) return [];

  // look for URI scheme according to RFC 2396
  var match = /([a-zA-Z0-9+\-.]+)\:/(query);
  var queryScheme = match != null? match[1]: null;
  var queryPath = match != null? query.substring(match[0].length): query;
  function testURI(uri) {
    if(!queryPath) return false;
    if(queryScheme != null && !uri.schemeIs(queryScheme)) return false;
    try { var path = uri.path.toLowerCase(); }
    catch(e) { return false; }
    return path.substring(0, queryPath.length) == queryPath;
  };

  // lax tel URI matching (URI later corrected by urifixup)
  var testTel = function() { return false; };
  // if query starts with "tel:" or with a numeric and it isn't part of an address
  if(/^tel\:/i.test(query) || (/^[0-9+]/.test(query) && !/[@\/]/.test(query))){
    var telQuery = query.replace(/^[^+0-9*#]+/, "");
    // return empty result if tel number is /empty/
    if(!telQuery || !/[1-9]/.test(telQuery)) return null;
    // count leading zeros
    var telQueryZeros = telQuery.search(/[^0]/);

    var testTel = function(uri) {
      if(!uri.schemeIs("tel")) return false;
      // strip visuals
      var number = decodeURI(uri.path).replace(/[^+0-9*#]/g, "");
      // leading + in query shall also match 00, 000 and so on
      if(telQuery.charAt(0) == '+') number = number.replace(/^00+/, "+");
      // remove leading + in number (49 also matches +49)
      else if(number.charAt(0) == '+') number = number.substring(1);
      // remove additional zeros (004 or 4 shall also match 0004)
      var zeros = number.search(/[^0]/);
      if(zeros > telQueryZeros) number = number.substring(zeros-telQueryZeros);

      return number.substring(0, telQuery.length) == telQuery;
    }
  }
  function testName(displayName) {
    var index = displayName.toLowerCase().indexOf(query);
    return index >= 0 && (index == 0 || /[^\w\xC0-\xFF]/.test(displayName.charAt(index-1)));
  };

  var items = this.items,
    result = new Result;
  for(var uriSpec in items) {
    var uriItems = items[uriSpec];
    for(var i = 0; i < uriItems.length; i++) {
      var item = uriItems[i];
      item.matched = false;
      if(testURI(item.uri) || testTel(item.uri) || testName(item.identity.displayName)) {
        item.matched = true;
        result.push(item);
      }
    }
  }
  return result.addresses;
};


function Session() {
  this.service = Components.classes["@mozilla.org/phone/address-lookup-service;1"].
    getService(Components.interfaces.phIAddressLookupService);
}
var session = Session.prototype = new Component("nsIAutoCompleteSession");

session.service = null;

session.onStartLookup = function(searchString, previousSearchResult, listener) {
  var lookupService = Components.classes["@mozilla.org/phone/address-lookup-service;1"].
    getService(Components.interfaces.phIAddressLookupService);
  var Status = Components.interfaces.nsIAutoCompleteStatus;

  var results = Components.classes["@mozilla.org/autocomplete/results;1"].
    createInstance(Components.interfaces.nsIAutoCompleteResults);
  var items = results.items;
  results.searchString = searchString;

  try { var addresses = lookupService.matchAddresses(searchString); }
  catch(e) {
    listener.onAutoComplete(null, Status.failed);
    return;
  }
  if(addresses == null) {
    listener.onAutoComplete(null, Status.ignored);
    return;
  }
  for(var i = 0, length = addresses.length; i < length; i++) {
    var address = addresses.queryElementAt(i, Components.interfaces.phIAddress),
      uri = address.uri;
    if(!uri.schemeIs("sip") && !uri.schemeIs("sips") && !uri.schemeIs("tel")) continue;
    var item = Components.classes["@mozilla.org/autocomplete/item;1"].
      createInstance(Components.interfaces.nsIAutoCompleteItem);
    item.value = address.uri.spec;
    item.comment = address.displayName;
    items.AppendElement(item);
  }
  listener.onAutoComplete(results, length? Status.matchFound: Status.noMatch);
};
session.onStopLookup = function() {
};
session.onAutoComplete = function(searchString, previousSearchResult, listener) {
};


var serviceComponent = new ComponentInfo(
  Service,
  "SIP Address Lookup Service",
  "@mozilla.org/phone/address-lookup-service;1",
  "2DCFE7A3-C85B-4d94-B99C-7C235E07C49F");

var sessionComponent = new ComponentInfo(
  Session,
  "SIP Auto Complete for Addresses",
  "@mozilla.org/autocompleteSession;1?type=phoneaddress",
  "4C7D42F5-AB6E-4e6a-B3E9-58715B10F433");

var module = new Module(serviceComponent, sessionComponent);
function NSGetModule() { return module; }
