Components.utils.importModule("chrome://phone/content/utils/class.js");

function getPhoneWindow(onLoad) {
  var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].
    getService(Components.interfaces.nsIWindowMediator);
  var window = windowManager.getMostRecentWindow("phone:call");
  if(window) {
    window.focus();
    if(onLoad != null) onLoad.call(window, null);
  }
  else {
    var windowWatcher = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].
      getService(Components.interfaces.nsIWindowWatcher);
    window = windowWatcher.openWindow(null, "chrome://phone/content/call.xul", "_blank", "chrome,extrachrome,dialog=no,all", null);
    if(onLoad != null) window.addEventListener("load", onLoad, false);
  }
  return window;
}

function getHeaderName(key) {
  var shortKey = key.charAt(0).toLowerCase();
  if(shortKey == 'f') return "From";
  else if(shortKey == 't') return "To";
  else if(shortKey == 's') return "Subject";
  else throw "Unknown header";
}

function Call() {
  this.listeners = [];
  this.headers = {};
}
var call = Call.prototype = new Component("phICall", "phISipProgressListener");

call.init = function(userAgent) {
  if(userAgent == null) {
    var manager = Components.classes["@mozilla.org/phone/sip-ua-manager;1"].
      getService(Components.interfaces.phISipUserAgentManager);
    // some day null should be replaced by a reference to phISipServer/nsIMsgIncomingServer
    userAgent = manager.getClient(null);
  }
  this.setHeader("From", userAgent.address.spec);
  this.userAgent = userAgent;
}

call.userAgent = null;
call.mediaSession = null;

call.listeners = null;
call.headerSink = null;

call.address = null;
call.uri getter = function() { return this.address? this.address.uri: null; };

call.dispatchListeners = function(/*...*/) {
  var propertyName = arguments[0];
  arguments[0] = this;
  var listeners = this.listeners;
  for(var i = 0; i < listeners.length; i++) {
    var listener = listeners[i];
    listener[propertyName].apply(listener, arguments);
  }
};
call.addProgressListener = function(listener) {
  this.listeners.push(listener);
};
call.removeProgressListener = function(listener) {
  var listeners = this.listeners;
  for(var i = 0; i < listeners.length; i++)
    if(listeners[i] == listener) delete listeners[i];
};

call.isCalling = false;
call.isLoading = false;
call.inviteConfirmed = false;
call.sessionEstablished = false;
call.terminated = false;

call.status = "";
call.title = "";
call.location = null;

call.loadAddress = function(address) {
  this.address = address;
  this.setHeader("To", address.spec);
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onStatusChange", this.status = "");
  this.dispatchListeners("onTitleChange", this.title = address.displayName);
  this.dispatchListeners("onLocationChange", this.location = address.uri.spec);
};
call.loadURI = function(uri) {
  var address = Components.classes["@mozilla.org/phone/address;1"].
    createInstance(Components.interfaces.phIAddress);
  address.uri = uri;
  this.loadAddress(address);
};
call.loadURISpec = function(uriSpec) {
  var fixup = Components.classes["@mozilla.org/phone/phone-urifixup;1"].
    getService(Components.interfaces.nsIURIFixup);
  try { var uri = fixup.createFixupURI(uriSpec, 0); }
  catch(e) { throw "Invalid URI"; }
  this.loadURI(uri);
};

call.headers = null;
call.setHeader = function(name, value) {
  name = getHeaderName(name);
  this.headers[name] = value;
  if(this.headerSink != null) this.headerSink.onHeader(name, value);
};
call.getHeader = function(name) {
  name = getHeaderName(name);
  if(!this.headers.hasOwnProperty(name)) return null;
  return this.headers[name];
};

// zapISipInviteMCH and zapISipDialogHandler implementation

function CallClient() {
  Call.apply(this, arguments);
}
var callClient = CallClient.prototype = new ComponentExtension(Call, "zapISipInviteMCH", "zapISipDialogHandler");

this.dialog = null;

callClient.dialogSpawned = function(methodClient, dialog) {
  dialog.setDialogHandler(this);
  this.dialog = dialog;
};
callClient.calling = function(methodClient, endpoint) {
  this.isCalling = true;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onInvite", null);
  this.dispatchListeners("onStatusChange", this.status = "Calling");
};
callClient.failureResponse = function() {
};
callClient.provisionalResponse = function() {
};
callClient.successResponse = function() {
  this.inviteConfirmed = true;
};
callClient.failure = function() {
  this.isLoading = false;
  this.isCalling = false;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onStatusChange", this.status = "Error");
};
callClient.success = function() {
  this.isLoading = false;
  this.isCalling = false;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onStatusChange", this.status = "Finished");
};
callClient.dialogConfirmed = function(dialog) {
};
callClient.dialogTerminated = function(dialog) {
};

callClient.invite = function() {
  this.isLoading = true;
  this.dispatchListeners("onStatusChange", this.status = "Connecting");

  var ioService = Components.classes["@mozilla.org/network/io-service;1"].
    getService(Components.interfaces.nsIIOService);

  // from here on we interoperate with the Zap interfaces
  var stack = this.userAgent.stack, sdpService = this.userAgent.sdpService;

  var address = this.address;
  // replace tel URIs according to RFC 3261
  // XXX done lazily here
  if(address.uri.schemeIs("tel")) {
    var sipURI = Components.classes["@mozilla.org/phone/sip-uri;1"].
      createInstance(Components.interfaces.nsIURI);
    var userURI = this.userAgent.address.uri;
    sipURI.spec = userURI.scheme+":"+encodeURIComponent(address.uri.path.toLowerCase())+"@"+userURI.hostPort;
    address.uri = sipURI;
  }

  var toAddress = stack.syntaxFactory.deserializeAddress(this.address.spec),
    fromAddress = stack.syntaxFactory.deserializeAddress(this.userAgent.address.spec);

  var fromURI = fromAddress.uri.QueryInterface(Components.interfaces.zapISipSIPURI);
  var offer = sdpService.createSessionDescription();
  // o=
  offer.username = fromURI.userinfo;
  offer.sessionID = "0";
  offer.sessionVersion = "0";
  offer.originAddressType = "IP4";
  offer.originAddress = fromURI.host;
  // s=
  offer.sessionName = " ";
  // c=
  offer.connection = sdpService.createConnection();
  offer.connection.addressType = "IP4";
  offer.connection.address = stack.hostAddress;
  // t=
  var time = sdpService.createTime();
  time.value = "0 0";
  offer.setTimes([time], 1);
  // m=
  var mediaDescription = sdpService.createMediaDescription();
  mediaDescription.media = "audio";
  mediaDescription.port = 0;//this.mediaSession.localRTPPort;
  mediaDescription.portCount = "";
  mediaDescription.protocol = "RTP/AVP";
  var fmt = sdpService.createMediaFormat();
  fmt.format = "97";
  mediaDescription.setFormats([fmt], 1);

  // a=
  var attribs = [];
  var attrib = sdpService.createAttrib();
  attrib.value = "rtpmap:97 speex/8000";
  attribs.push(attrib);

  mediaDescription.setAttribs(attribs, attribs.length);

  offer.setMediaDescriptions([mediaDescription], 1);

  var request = stack.formulateInviteRequest(toAddress, fromAddress);
  request.setContent("application", "sdp", offer.serialize());

  /*...*/
  this.sessionEstablished = true;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onACK", null);

  var inviteMC = stack.createInviteMC(request, this);
  inviteMC.execute();  
};
callClient.bye = function() {
  if(!this.dialog) throw "No spawned dialog to quit";
  this.dialog.bye();
  this.dispatchListeners("onBye", null);
};

// zapISipInviteMSH implementation

function CallServer() {
  Call.apply(this, arguments);
}
var callServer = CallServer.prototype = new ComponentExtension(Call, "zapISipInviteMSH");

callServer.dialog = null;
callServer.request = null;

callServer.invite = function(methodServer, request) {
  this.methodServer = methodServer;
  this.request = request;

  // creating a phIAddress from a zapISipZapAddress
  // XXX someday the ph and zap address interfaces may be joined
  var address = this.address = Components.classes["@mozilla.org/phone/address;1"].
    createInstance(Components.interfaces.phIAddress);
  address.spec = request.getFromHeader().address.serialize();

  this.setHeader("To", request.getToHeader().serialize());
  this.isCalling = true;
  this.isLoading = true;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onInvite", null);
  this.dispatchListeners("onStatusChange", this.status = "Connecting");
  this.dispatchListeners("onTitleChange", this.title = address.displayName);
  this.dispatchListeners("onLocationChange", this.location = address.uri.spec);

  // parse the INVITE's sdp to see if it is acceptable:
  try { var sd = this.userAgent.sdpService.deserializeSessionDescription(request.body); }
  catch(e) {
    this.decline();
    return;
  }
  try {
    var mediaDescriptions = sd.getMediaDescriptions({});
    // XXX assuming that there is only one media description for the moment:
    
    this.remoteRtpPort = mediaDescriptions[0].port;        
    var connection = mediaDescriptions[0].connection;
    if (!connection)
      connection = sd.connection;
    this.remoteMediaHost = connection.address;

    // find media description that contains an a=rtpmap:xx speex/8000
    // attribute to determine payload type:
    var attribs = mediaDescriptions[0].getAttribs({});
    var match;
    for (var i=0,l=attribs.length; i<l; ++i)
      if ((match = /rtpmap:(\d+) speex\/8000/(attribs[i].value))) {
        this.remotePayloadFormat = match[1];
        break;
      }
  }
  catch(e) {
    this.decline();
    return;
  }
  // if we get to here, the remote host's offering is acceptable
  this.methodServer = methodServer;
  this.request = request;

//  window.confirm("Accept call from "+address.displayName+"?");

  var alertService = Components.classes["@mozilla.org/alerts-service;1"].
    getService(Components.interfaces.nsIAlertsService);
  alertService.showAlertNotification(
    "chrome://phone/skin/icons/phone-settings.png",
    "Receiving Call from "+address.displayName,
    "Click here to accept the call now",
    true,
    "",
    {
      call: this,
      observe: function(subject, topic, data) {
        if(topic == "alertclickcallback") {
          this.call.accept();
          this.accepted = true;
        }
      }
    });
};
callServer.dialogSpawned = function(methodServer, dialog) {
  dialog.setDialogHandler(this);
  this.dialog = dialog;
};
callServer.failure = function(methodServer) {
  this.isLoading = false;
  this.isCalling = false;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onStatusChange", this.status = "Error");
};
callServer.success = function(methodServer, dialog) {
  this.isLoading = false;
  this.isCalling = false;
  this.dispatchListeners("onStateChange", null);
  this.dispatchListeners("onStatusChange", this.status = "Finished");
};
callServer.dialogConfirmed = function(dialog) {
};
callServer.dialogTerminated = function(dialog) {
};

callServer.decline = function() {
  var r = this.userAgent.stack.formulateResponse("603", "Decline", this.request, true);
  this.methodServer.sendResponse(r);
};
callServer.bye = function() {
  if(this.dialog != null) this.dialog.bye();
  this.dispatchListeners("onBye", null);
};
callServer.element = null;
callServer.accept = function() {
  if(this.mediaSession == null)
    this.mediaSession = Components.classes["@mozilla.org/zap/mediasession;1"].
      createInstance(Components.interfaces.zapIMediaSession);

  var call = this;
  if(this.element == null) {
    getPhoneWindow(function() {
      this.document.getElementById('tabphone').acceptCall(call);
    });
  }
  return;
  var r = this.methodServer.formulate200Response();

  // set media offer:
  var fromAddress = this.userAgent.stack.syntaxFactory.deserializeAddress(this.userAgent.address.spec);
  var fromURI = fromAddress.uri.QueryInterface(Components.interfaces.zapISipSIPURI);
  var offer = this.userAgent.sdpService.createSessionDescription();
  // o=
  offer.username = fromURI.userinfo;
  offer.sessionID = "0";
  offer.sessionVersion = "0";
  offer.originAddressType = "IP4";
  offer.originAddress = fromURI.host;
  // s=
  offer.sessionName = " ";
  // c=
  offer.connection = this.userAgent.sdpService.createConnection();
  offer.connection.addressType = "IP4";
  offer.connection.address = this.userAgent.stack.hostAddress;
  // t=
  var time = this.userAgent.sdpService.createTime();
  time.value = "0 0";
  offer.setTimes([time], 1);
  // m=
  var mediaDescription = this.userAgent.sdpService.createMediaDescription();
  mediaDescription.media = "audio";
  mediaDescription.port = this.mediaSession.localRTPPort;
  mediaDescription.portCount = "";
  mediaDescription.protocol = "RTP/AVP";
  var fmt = this.userAgent.sdpService.createMediaFormat();
  fmt.format = "97";
  mediaDescription.setFormats([fmt], 1);
  
  // a=
  var attribs = [];
  var attrib = this.userAgent.sdpService.createAttrib();
  attrib.value = "rtpmap:97 speex/8000";
  attribs.push(attrib);
  
  mediaDescription.setAttribs(attribs, attribs.length);
  
  offer.setMediaDescriptions([mediaDescription], 1);
  
  r.setContent("application", "sdp", offer.serialize());
  this.methodServer.sendResponse(r);
};

var clientComponent = new ComponentInfo(
  CallClient,
  "Received Call",
  "@mozilla.org/phone/call;1?type=client",
  "BFA9216F-C93F-42b2-88BF-839C12B910EB");
var serverComponent = new ComponentInfo(
  CallServer,
  "Sent Call",
  "@mozilla.org/phone/call;1?type=server",
  "C096D01D-19A4-4f24-93A2-DED6DA7E8217");

var module = new Module(clientComponent, serverComponent);
function NSGetModule() { return module; }
