Components.utils.importModule("chrome://phone/content/utils/class.js");

function unq(str) {
  return str.substring(1, str.length-1);
}
function q(str) {
  if(str.indexOf('"') >= 0) throw "String must not contain quotes";
  return '"'+str+'"';
}

function Authenticator() { }
var authenticator = Authenticator.prototype = new Component("zapISipAuthenticator");

authenticator.stack = null;

authenticator.init = function(stack) {
  this.stack = stack;
};

authenticator.generateCredentials = function(
  request,
  response,
  proxyAuth,
  username,
  passwd
) {
  var hasher = Components.classes["@mozilla.org/security/hash;1"].
    createInstance(Components.interfaces.nsICryptoHash);

  function MD5(str) {
    if(!str.length) return "d41d8cd98f00b204e9800998ecf8427e";
    hasher.init(Components.interfaces.nsICryptoHash.MD5);
    var stream = Components.classes["@mozilla.org/io/string-input-stream;1"].
      createInstance(Components.interfaces.nsIStringInputStream);
    stream.setData(str, str.length);
    hasher.updateFromStream(stream, str.length);
    var bin = hasher.finish(false);

    for(var hex = "", i = 0; i < bin.length; i++) {
      var code = bin.charCodeAt(i);
      hex += "0123456789abcdef"[code >> 4] + "0123456789abcdef"[code & 15];
    }
    return hex;
  }

  var challengeName = proxyAuth? "Proxy-Authenticate": "WWW-Authenticate";
  var credentialsName = proxyAuth? "Proxy-Authorization": "Authorization";

  // see RFC 2617 section 3.2
  // XXX support multiple WWW-Authenticate headers
  var challenge = response.getSingleHeader(challengeName);
  if(challenge == null) throw "No authenticate header found";
  challenge = challenge.QueryInterface(Components.interfaces.zapISipAuthHeader);
  var credentials = this.stack.syntaxFactory.createHeader(credentialsName, challenge.value).
    QueryInterface(Components.interfaces.zapISipAuthHeader);

  var scheme = challenge.scheme.toLowerCase();
  if(scheme == "digest") {
    var H, A1, A2;

    if(credentials.hasAuthParam("username") && unq(credentials.getAuthParam("username")) != username)
      throw "Login for different user required by UAS";
    credentials.setAuthParam("username", q(username));
    if(!credentials.hasAuthParam("uri"))
      credentials.setAuthParam("uri", q(request.requestURI.serialize()));
    var qop_value = challenge.getAuthParam("qop"),
      algorithm_value = challenge.getAuthParam("algoritm"),
      username_value = credentials.getAuthParam("username"),
      digest_uri_value = credentials.getAuthParam("uri"),
      realm_value = challenge.getAuthParam("realm");
    if(algorithm_value == "MD5" || algorithm_value == null) {
      H = MD5;
      A1 = unq(username_value)+":"+unq(realm_value)+":"+passwd;
    }
    else {
      throw "Algorithm not supported";
    }

    if(qop_value == "auth" || qop_value == null) {
      A2 = request.method+":"+unq(digest_uri_value);
    }
    else if(qop_value == "auth-int") {
      A2 = request.method+":"+unq(digest_uri_value)+":"+H(request.body);
    }
    else {
      throw "Option not supported";
    }

    if(qop_value == "auth" || qop_value == "auth-int") {
      var nonce_value = challenge.getAuthParam("nonce"),
        nc_value = challenge.getAuthParam("nonce-count"),
        cnonce_value = challenge.getAuthParam("cnonce");
      var request_digest = '"'+H(H(A1)+":"+unq(nonce_value)+":"+nc+":"+unq(cnonce_value)+":"+unq(qop_value)+":"+H(A2))+'"';
    }
    else if(qop_value == null) {
      var nonce_value = challenge.getAuthParam("nonce");
      var request_digest = '"'+H(H(A1)+":"+unq(nonce_value)+":"+H(A2))+'"';
    }

    credentials.setAuthParam("response", request_digest);
  }
  else {
    throw "Scheme not supported";
  }

  request.removeHeaders(credentialsName);
  request.appendHeader(credentials);
};

authenticator.uri = null;

var component = new ComponentInfo(
  Authenticator,
  "SIP Authenticator",
  "@mozilla.org/network/sip-authenticator;1",
  "F5D801FA-5BEA-4ccb-B91F-42950237F823");

var module = new Module(component);
function NSGetModule() { return module; }
