Components.utils.importModule("chrome://phone/content/utils/class.js");

function CallParams() { }
var callParams = CallParams.prototype = new Component("phISipCallParams");
callParams.address = null;

function ChatParams() { }
var chatParams = ChatParams.prototype = new Component("phISipChatParams");
chatParams.address = null;

function ProtocolHandler(scheme, defaultPort) {
  this.scheme = scheme;
  this.defaultPort = defaultPort != null? defaultPort: -1;
}
var protocolHandler = ProtocolHandler.prototype = new Component("nsIProtocolHandler");

protocolHandler.scheme = null;
protocolHandler.defaultPort = null;
protocolHandler.protocolFlags = Components.interfaces.nsIProtocolHandler.URI_NORELATIVE | Components.interfaces.nsIProtocolHandler.URI_NOAUTH;

protocolHandler.allowPort = function(port, scheme) { return false; };

protocolHandler.newURI = function(spec, originCharset, baseURI) {
  var uri = Components.classes["@mozilla.org/phone/sip-uri;1"].createInstance(Components.interfaces.nsIURI);
  uri.originCharset = originCharset;
  uri.spec = spec;
  return uri;
};

protocolHandler.newChannel = function(uri) {
  var params = Components.classes["@mozilla.org/phone/callparams;1"].
    createInstance(Components.interfaces.phISipCallParams);
  var address = Components.classes["@mozilla.org/phone/address;1"].
    createInstance(Components.interfaces.phIAddress);
  address.uri = uri;
  params.address = address;

  var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].
    getService(Components.interfaces.nsIWindowMediator);
  var window = windowManager.getMostRecentWindow("phone:call");
  if(window) {
    window.focus();
    window.loadURI(uri.spec);
  }
  else {
    var windowWatcher = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].
      getService(Components.interfaces.nsIWindowWatcher);
    window = windowWatcher.openWindow(null, "chrome://phone/content/call.xul", "_blank", "chrome,extrachrome,dialog=no,all", params);
  }

  var channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].
    getService(Components.interfaces.nsIInputStreamChannel);
  channel.contentType = "application/x-sip";
  channel.setURI(uri);
  return channel;
};

function SIPProtocolHandler() {}
SIPProtocolHandler.prototype = new ProtocolHandler("sip", 5060);

function SIPSProtocolHandler() {}
SIPSProtocolHandler.prototype = new ProtocolHandler("sips", 5061);

function TELProtocolHandler() {}
TELProtocolHandler.prototype = new ProtocolHandler("tel");
TELProtocolHandler.prototype.newURI = function(spec, originCharset, baseURI) {
  var uri = Components.classes["@mozilla.org/network/simple-uri;1"].createInstance(Components.interfaces.nsIURI);
  uri.originCharset = originCharset;
  uri.spec = spec;
  return uri;
};

var callParamsComponent = new ComponentInfo(
  CallParams,
  "SIP Call Parameters",
  "@mozilla.org/phone/callparams;1",
  "30508F1C-9F01-4ecb-A35A-A861A7F43FF3");

var chatParamsComponent = new ComponentInfo(
  ChatParams,
  "SIP Chat Parameters",
  "@mozilla.org/phone/chatparams;1",
  "8AADE147-38A3-4afb-A55D-223BB69BABCA");

var sipComponent = new ComponentInfo(
  SIPProtocolHandler,
  "SIP Protocol Handler",
  "@mozilla.org/network/protocol;1?name=sip",
  "789409b9-2e3b-4682-a5d1-71ca80a76456");

var sipsComponent = new ComponentInfo(
  SIPSProtocolHandler,
  "SIPS Protocol Handler",
  "@mozilla.org/network/protocol;1?name=sips",
  "A428F471-DE1E-4458-98BD-73774C162C02");

var telComponent = new ComponentInfo(
  TELProtocolHandler,
  "TEL Protocol Handler",
  "@mozilla.org/network/protocol;1?name=tel",
  "B9D12E37-C5F1-4a9e-8AB5-9F89B2E472BA");

var module = new Module(callParamsComponent, chatParamsComponent, sipComponent, sipsComponent, telComponent);

module.registerSelf = function(compMgr, fileSpec, location, type) {
  Module.prototype.registerSelf.call(this, compMgr, fileSpec, location, type);

// XXX why doesn't this have an effect?
  var prefs = Components.classes["@mozilla.org/preferences-service;1"].
    getService(Components.interfaces.nsIPrefBranch);
  prefs.setBoolPref("network.protocol-handler.expose.sip", true);
  prefs.setBoolPref("network.protocol-handler.expose.sips", true);
  prefs.setBoolPref("network.protocol-handler.expose.tel", true);
}
module.unregisterSelf = function(compMgr, fileSpec, location) {
  Module.prototype.unregisterSelf.call(this, compMgr, fileSpec, location);

// XXX why doesn't this have an effect?
  var prefs = Components.classes["@mozilla.org/preferences-service;1"].
    getService(Components.interfaces.nsIPrefBranch);
  prefs.clearUserPref("network.protocol-handler.expose.sip");
  prefs.clearUserPref("network.protocol-handler.expose.sips");
  prefs.clearUserPref("network.protocol-handler.expose.tel");
}

function NSGetModule() { return module; }